@@ -8,73 +8,96 @@ import json
 import math
 import random
 import signal
+import threading
 import time
 import traceback
 
+
 import gevent.pool
 import multiprocessing as mp
 import numpy as np
 import pathos.pools
 from tqdm import tqdm
 
-from cloudvolume.threaded_queue import ThreadedQueue
-from cloudvolume.lib import yellow, scatter
+from .threaded_queue import ThreadedQueue
+from .lib import yellow, scatter, sip, toiter
 
-from .aws_queue_api import AWSTaskQueueAPI
-from .registered_task import RegisteredTask, deserialize
-from .scheduler import schedule_green_jobs, schedule_threaded_jobs
+from .aws_queue_api import AWSTaskQueueAPI, AWS_BATCH_SIZE
+from .file_queue_api import FileQueueAPI
+from .paths import extract_path, mkpath
+from .registered_task import RegisteredTask, deserialize, totask, totaskid
+from .scheduler import schedule_jobs
 from .secrets import (
-  PROJECT_NAME, QUEUE_NAME, QUEUE_TYPE,
   AWS_DEFAULT_REGION
 )
 
-AWS_BATCH_SIZE = 10 
+def totalfn(iterator, total):
+  if total is not None:
+    return total
+  try:
+    return len(iterator)
+  except TypeError:
+    return None
 
-def totask(task):
-  if isinstance(task, RegisteredTask):
-    return task
-
-  taskobj = deserialize(task['payload'])
-  taskobj._id = task['id']
-  return taskobj
-
-class QueueEmpty(LookupError):
-  def __init__(self):
-    super(QueueEmpty, self).__init__('Queue Empty')
+class QueueEmptyError(LookupError):
+  pass
 
 LEASE_SECONDS = 300
 
-class SuperTaskQueue(object):
+class TaskQueue(object):
+  """
+  The standard usage is that a client calls lease to get the next available task,
+  performs that task, and then calls task.delete on that task before the lease expires.
+  If the client cannot finish the task before the lease expires,
+  and has a reasonable chance of completing the task,
+  it should call task.update before the lease expires.
+  If the client completes the task after the lease has expired,
+  it still needs to delete the task. 
+  Tasks should be designed to be idempotent to avoid errors 
+  if multiple clients complete the same task.
+  """
   def __init__(
-    self, queue_name=QUEUE_NAME, queue_server=QUEUE_TYPE, 
-    region=None, qurl=None, n_threads=40, project=PROJECT_NAME
+    self, qurl, n_threads=40, 
+    green=False, progress=True, 
+    **kwargs
   ):
-    self._project = project
-    self._region = region
-    self._queue_name = queue_name
-    self._queue_server = queue_server
-    self._qurl = qurl
-    self._api = self._initialize_interface()
+    self.qurl = qurl
+    self.path = extract_path(qurl)
+    self.api = self.initialize_api(self.path, kwargs)
+    self.n_threads = n_threads
+    self.green = bool(green)
+    self.progress = bool(progress),
+    self.kwargs = kwargs
 
-  # This is key to making sure threading works. Don't refactor this method away.
-  def _initialize_interface(self):
-    server = self._queue_server.lower()
-    if server in ('pull-queue', 'google'):
-      return NotImplementedError("Google Cloud Tasks are not supported at this time.")
-    elif server in ('sqs', 'aws'):
-      qurl = self._qurl if self._qurl else self._queue_name
-      region = self._region if self._region else AWS_DEFAULT_REGION
-      return AWSTaskQueueAPI(qurl=qurl, region_name=region)
+    if self.green:
+      self.check_monkey_patch_status()
+
+  @property
+  def qualified_path(self):
+    return mkpath(self.path)
+
+  def initialize_api(self, path, kwargs):
+    if path.protocol == 'sqs':
+      return AWSTaskQueueAPI(qurl=path.path, region_name=kwargs.get('region', AWS_DEFAULT_REGION))
+    elif path.protocol == 'fq':
+      return FileQueueAPI(path.path)
     else:
-      raise NotImplementedError('Unknown server ' + self._queue_server)
+      raise ValueError('Unsupported protocol ' + str(self.path.protocol))
 
-  @property
-  def queue_name(self):
-    return self._queue_name
-  
-  @property
-  def queue_server(self):
-    return self._queue_server
+  def check_monkey_patch_status(self):
+    import gevent.monkey
+    if not gevent.monkey.is_module_patched("socket"):
+      print(yellow("""
+    Green threads require monkey patching the standard library 
+    to use a non-blocking network socket call.
+
+    Please place the following lines at the beginning of your
+    program. `thread=False` is there because sometimes this
+    causes hanging in multiprocessing.
+
+    import gevent.monkey
+    gevent.monkey.patch_all(thread=False)
+        """))
 
   @property
   def enqueued(self):
@@ -88,113 +111,134 @@ class SuperTaskQueue(object):
     
     Returns: (int) number of tasks in cloud queue
     """
-    return self._api.enqueued
+    return self.api.enqueued
 
-  def status(self):
-    """
-    Gets information about the TaskQueue
-    """
-    return self._api.get(getStats=True)
+  @property
+  def inserted(self):
+    return self.api.inserted
 
-  def get_task(self, tid):
-    """
-    Gets the named task in the TaskQueue. 
-    tid is a unique string Google provides 
-    e.g. '7c6e81c9b7ab23f0'
-    """
-    return self._api.get(tid)
+  @property
+  def completed(self):
+    return self.api.completed
+
+  def is_empty(self):
+    return self.api.is_empty()
+
+  # def status(self):
+  #   """
+  #   Gets information about the TaskQueue
+  #   """
+  #   return self.api.get(getStats=True)
 
   def list(self):
     """
     Lists all non-deleted Tasks in a TaskQueue, 
-    whether or not they are currently leased, up to a maximum of 100.
+    whether or not they are currently leased, 
+    up to a maximum of 100.
     """
-    return [ totask(x) for x in self._api.list() ]
+    return [ totask(x) for x in iter(self.api) ]
 
-  def insert_all(self, tasks, delay_seconds=0, total=None, parallel=1):
-    if total is None:
-      try:
-        total = len(tasks)
-      except TypeError:
-        pass
+  def insert(self, tasks, delay_seconds=0, total=None, parallel=1):
+    tasks = toiter(tasks)
+    total = totalfn(tasks, total)
 
-    if parallel not in (1, False):
-      multiprocess_upload(self.__class__, self.queue_name, tasks, parallel=parallel)
-    else:
-      self._insert_all(tasks, delay_seconds, total)
-
-  def _gen_insert_all_tasks(
-      self, tasks, 
-      batch_size=AWS_BATCH_SIZE, delay_seconds=0, total=None
-    ):
-     
-    def genbatches(itr):
-      while True:
-        batch = []
-        try:
-          for i in range(AWS_BATCH_SIZE):
-            batch.append(next(itr))
-        except StopIteration:
-          pass
-
-        if len(batch) == 0:
-          break
-
-        yield batch
+    if parallel not in (1, False) and total is not None and total > 1:
+      multiprocess_upload(self.__class__, mkpath(self.path), tasks, parallel=parallel)
+      return
 
+    try:
+      batch_size = self.api.batch_size
+    except:
+      batch_size = 1
+    
     bodies = (
       {
         "payload": task.payload(),
-        "queueName": self._queue_name,
-        "groupByTag": True,
-        "tag": task.__class__.__name__
+        "queueName": self.path.path,
       } 
-
       for task in tasks
     )
 
-    def cloud_insertion(batch):
-      self._api.insert(batch, delay_seconds)
+    ct = [ 0 ] # using a list instead of a raw number to use pass-by-reference
+    ct_lock = threading.Lock()
+    def insertfn(batch, ct):
+      try:
+        incr = self.api.insert(batch, delay_seconds) 
+      finally:
+        with ct_lock:
+          ct[0] += incr
+    
+    schedule_jobs(
+      fns=( partial(insertfn, batch, ct) for batch in sip(bodies, batch_size) ),
+      concurrency=self.n_threads,
+      progress='Inserting',
+      total=total,
+      green=self.green,
+      batch_size=batch_size,
+    )
 
-    return ( partial(cloud_insertion, batch) for batch in genbatches(bodies) )
+    self.api.add_insert_count(ct[0])
 
-  def renew_lease(self, task, seconds):
+  def insert_all(self, *args, **kwargs):
+    """For backwards compatibility."""
+    return self.insert(*args, **kwargs)
+
+  def rezero(self):
+    """Resets statistic counters such as completions and insertions to zero."""
+    self.api.rezero()
+
+  def renew(self, task, seconds):
     """Update the duration of a task lease."""
-    return self._api.renew_lease(task, seconds)
+    return self.api.renew_lease(task, seconds)
 
-  def cancel_lease(self, task):
-    return self._api.cancel_lease(task)
+  def cancel(self, task):
+    return self.api.cancel_lease(task)
 
-  def lease(self, seconds=600, num_tasks=1, tag=None):
+  def release_all(self):
+    return self.api.release_all()
+
+  def lease(self, seconds=600, num_tasks=1):
     """
     Acquires a lease on the topmost N unowned tasks in the specified queue.
     Required query parameters: leaseSecs, numTasks
     """
-    tag = tag if tag else None
-    tasks = self._api.lease(
-      numTasks=num_tasks, 
-      seconds=seconds,
-      groupByTag=(tag is not None),
-      tag=tag,
-    )
+    if num_tasks <= 0:
+      raise ValueError("num_tasks must be > 0. Got: " + str(num_tasks))
+    if seconds < 0:
+      raise ValueError("lease seconds must be >= 0. Got: " + str(seconds))
+
+    tasks = self.api.lease(seconds, num_tasks)
 
     if not len(tasks):
-      raise QueueEmpty
+      raise QueueEmptyError()
 
-    task = tasks[0]
-    return totask(task)
+    if num_tasks == 1:
+      return totask(tasks[0])
+    else:
+      return [ totask(task) for task in tasks ]
 
-  def patch(self):
-    """
-    Update tasks that are leased out of a TaskQueue.
-    Required query parameters: newLeaseSeconds
-    """
-    raise NotImplemented
+  def delete(self, task_id, total=None, tally=False):
+    """Deletes a task from a TaskQueue."""
+    task_id = toiter(task_id)
+    total = totalfn(task_id, total)
+
+    def deltask(tid):
+      self.api.delete(totaskid(tid))
+      if tally:
+        self.api.tally()
+
+    schedule_jobs(
+      fns=( partial(deltask, tid) for tid in task_id ),
+      concurrency=self.n_threads,
+      progress=('Deleting' if self.progress else None),
+      total=total,
+      green=self.green,
+    )
 
   def purge(self):
     """Deletes all tasks in the queue."""
     try:
-      return self._api.purge()
+      return self.api.purge()
     except AttributeError:
       while True:
         lst = self.list()
@@ -207,17 +251,17 @@ class SuperTaskQueue(object):
       return self
 
   def poll(
-    self, lease_seconds=LEASE_SECONDS, tag=None, 
+    self, lease_seconds=LEASE_SECONDS,  
     verbose=False, execute_args=[], execute_kwargs={}, 
-    stop_fn=None, backoff_exceptions=[], min_backoff_window=30, 
-    max_backoff_window=120, before_fn=None, after_fn=None
+    stop_fn=None, backoff_exceptions=[], min_backoff_window=1, 
+    max_backoff_window=120, before_fn=None, after_fn=None,
+    tally=False
   ):
     """
     Poll a queue until a stop condition is reached (default forever). Note
     that this function is not thread safe as it requires a global variable
     to intercept SIGINT.
     lease_seconds: each task should be leased for this many seconds
-    tag: if specified, query for only tasks that match this tag
     execute_args / execute_kwargs: pass these arguments to task execution
     backoff_exceptions: A list of exceptions that instead of causing a crash,
       instead cause the polling to back off for an increasing exponential 
@@ -228,10 +272,14 @@ class SuperTaskQueue(object):
       in seconds.
     stop_fn: A boolean returning function that accepts no parameters. When
       it returns True, the task execution loop will terminate. It is evaluated
-      once after every task.
+      once after every task. If you provide the arguments `executed` (tasks completed) 
+      `tries` (current attempts at fetching a task), or `previous_execution_time` (time in
+      seconds to run the last task) they will be dependency injected.
     before_fn: Pass task pre-execution.
     after_fn: Pass task post-execution.
     verbose: print out the status of each step
+    tally: contribute each completed task to a completions counter if supported.
+
     Return: number of tasks executed
     """
     global LOOP
@@ -267,7 +315,7 @@ class SuperTaskQueue(object):
     executed = 0
 
     backoff = False
-    backoff_exceptions = tuple(list(backoff_exceptions) + [ QueueEmpty ])
+    backoff_exceptions = tuple(list(backoff_exceptions) + [ QueueEmptyError ])
 
     before_fn = before_fn or (lambda x: x)
     after_fn = after_fn or (lambda x: x)
@@ -283,7 +331,7 @@ class SuperTaskQueue(object):
         time_delta = time.time() - time_start
         executed += 1
         printv("Delete enqueued task...")
-        self.delete(task)
+        self.delete(task, tally=tally)
         printv('INFO', task , "succesfully executed in {:.2f} sec.".format(time_delta))
         after_fn(task)
         tries = 0
@@ -292,8 +340,17 @@ class SuperTaskQueue(object):
       except Exception as e:
         printv('ERROR', task, "raised {}\n {}".format(e , traceback.format_exc()))
         raise # this will restart the container in kubernetes
-        
-      if stop_fn():
+      
+      varnames = stop_fn.__code__.co_varnames
+      stop_fn_bound = stop_fn
+      if 'executed' in varnames:
+        stop_fn_bound = partial(stop_fn_bound, executed=executed)
+      if 'tries' in varnames:
+        stop_fn_bound = partial(stop_fn_bound, tries=tries)
+      if 'previous_execution_time' in varnames:
+        stop_fn_bound = partial(stop_fn_bound, previous_execution_time=time_delta) 
+
+      if stop_fn_bound():
         break
 
       if backoff:
@@ -310,253 +367,6 @@ class SuperTaskQueue(object):
     while self.enqueued > 0:
       time.sleep(interval_sec)
 
-class TaskQueue(SuperTaskQueue, ThreadedQueue):
-  """
-  The standard usage is that a client calls lease to get the next available task,
-  performs that task, and then calls task.delete on that task before the lease expires.
-  If the client cannot finish the task before the lease expires,
-  and has a reasonable chance of completing the task,
-  it should call task.update before the lease expires.
-  If the client completes the task after the lease has expired,
-  it still needs to delete the task. 
-  Tasks should be designed to be idempotent to avoid errors 
-  if multiple clients complete the same task.
-  """
-  def __init__(
-    self, queue_name=QUEUE_NAME, queue_server=QUEUE_TYPE, 
-    region=None, qurl=None, n_threads=40, project=PROJECT_NAME
-  ):
-
-    SuperTaskQueue.__init__(
-      self, queue_name, queue_server, 
-      region, qurl, n_threads, project
-    )
-    ThreadedQueue.__init__(self, n_threads) # creates self._queue
-    
-  def insert(self, task, args=[], kwargs={}, delay_seconds=0):
-    """
-    Insert a task into an existing queue.
-    """
-    body = {
-      "payload": task.payload(),
-      "queueName": self._queue_name,
-      "groupByTag": True,
-      "tag": task.__class__.__name__
-    }
-
-    def cloud_insertion(api):
-      api.insert(body, delay_seconds)
-
-    if len(self._threads):
-      self.put(cloud_insertion)
-    else:
-      cloud_insertion(self._api)
-
-    return self
-
-  def _insert_all(self, tasks, delay_seconds=0, total=None):
-    batch_size = AWS_BATCH_SIZE
-    fns = self._gen_insert_all_tasks(
-      tasks, batch_size, delay_seconds, total
-    )
-
-    schedule_threaded_jobs(
-      fns=fns,
-      concurrency=20,
-      progress='Inserting',
-      total=total,
-      batch_size=batch_size,
-    )
-
-  def acknowledge(self, task_id):
-    if isinstance(task_id, RegisteredTask):
-      task_id = task_id.id
-
-    def cloud_delete(api):
-      api.acknowledge(task_id)
-
-    if len(self._threads):
-      self.put(cloud_delete)
-    else:
-      cloud_delete(self._api)
-
-    return self
-
-  def delete(self, task_id):
-    """Deletes a task from a TaskQueue."""
-    if isinstance(task_id, RegisteredTask):
-      task_id = task_id.id
-
-    def cloud_delete(api):
-      api.delete(task_id)
-
-    if len(self._threads):
-      self.put(cloud_delete)
-    else:
-      cloud_delete(self._api)
-
-    return self
-
-class GreenTaskQueue(SuperTaskQueue):
-  """
-  The standard usage is that a client calls lease to get the next available task,
-  performs that task, and then calls task.delete on that task before the lease expires.
-  If the client cannot finish the task before the lease expires,
-  and has a reasonable chance of completing the task,
-  it should call task.update before the lease expires.
-
-  If the client completes the task after the lease has expired,
-  it still needs to delete the task. 
-
-  Tasks should be designed to be idempotent to avoid errors 
-  if multiple clients complete the same task.
-  """
-  def __init__(
-    self, queue_name=QUEUE_NAME, queue_server=QUEUE_TYPE, 
-    region=None, qurl=None, n_threads=40, project=PROJECT_NAME
-  ):
-    n_threads = max(n_threads, 1)
-    SuperTaskQueue.__init__(
-      self, queue_name, queue_server, 
-      region, qurl, n_threads, project
-    )
-    self._pool = gevent.pool.Pool(n_threads)
-    self.check_monkey_patch_status()
-
-
-  def check_monkey_patch_status(self):
-    import gevent.monkey
-    if not gevent.monkey.is_module_patched("socket"):
-      print(yellow("""
-    GreenTaskQueue uses green threads. This requires
-    monkey patching the standard library to use a cooperative 
-    non-blocking threading model.
-
-    Please place the following lines at the beginning of your
-    program. `thread=False` is there because sometimes this
-    causes hanging in multiprocessing.
-
-    import gevent.monkey
-    gevent.monkey.patch_all(thread=False)
-        """))
-
-  def insert(self, task, args=[], kwargs={}, delay_seconds=0):
-    """
-    Insert a task into an existing queue.
-    """
-    body = {
-      "payload": task.payload(),
-      "queueName": self._queue_name,
-      "groupByTag": True,
-      "tag": task.__class__.__name__
-    }
-
-    def cloud_insertion():
-      self._api.insert(body, delay_seconds)
-
-    self._pool.spawn(cloud_insertion)
-
-    return self
-
-  def _insert_all(self, tasks, delay_seconds=0, total=None):
-    batch_size = AWS_BATCH_SIZE
-    fns = self._gen_insert_all_tasks(
-      tasks, batch_size, delay_seconds, total
-    )
-
-    schedule_green_jobs(
-      fns=fns,
-      concurrency=20,
-      progress='Inserting',
-      total=total,
-      batch_size=batch_size,
-    )
-
-  def status(self):
-    """
-    Gets information about the TaskQueue
-    """
-    return self._api.get(getStats=True)
-
-  def get_task(self, tid):
-    """
-    Gets the named task in the TaskQueue. 
-    tid is a unique string Google provides 
-    e.g. '7c6e81c9b7ab23f0'
-    """
-    return self._api.get(tid)
-
-  def list(self):
-    """
-    Lists all non-deleted Tasks in a TaskQueue, 
-    whether or not they are currently leased, up to a maximum of 100.
-    """
-    return [ totask(x) for x in self._api.list() ]
-
-  def acknowledge(self, task_id):
-    if isinstance(task_id, RegisteredTask):
-      task_id = task_id.id
-
-    def cloud_delete():
-      self._api.acknowledge(task_id)
-
-    self._pool.spawn(cloud_delete)
-
-    return self
-
-  def delete(self, task_id):
-    """Deletes a task from a TaskQueue."""
-    if isinstance(task_id, RegisteredTask):
-      task_id = task_id.id
-
-    def cloud_delete():
-      self._api.delete(task_id)
-
-    self._pool.spawn(cloud_delete)
-
-    return self
-
-  def wait(self):
-    self._pool.join()
-
-  def __enter__(self):
-    return self
-
-  def __exit__(self, exception_type, exception_value, traceback):
-    self._pool.join()
-
-class MockTaskQueue(object):
-  def __init__(self, *args, **kwargs):
-    self.progress = False 
-    if 'progress' in kwargs:
-      self.progress = bool(kwargs['progress'])
-
-  def insert(self, task, args=[], kwargs={}):
-    task = {
-      'payload': task.payload(),
-      'id': -1,
-    }
-    task = totask(task) # to ensure conformity with TaskQueue
-    task.execute(*args, **kwargs)
-    del task
-
-  def insert_all(
-    self, tasks, args=[], kwargs={}, 
-    delay_seconds=0, total=None,
-    parallel=1
-  ):
-    for task in tqdm(tasks, disable=(not self.progress)):
-      self.insert(task, args=args, kwargs=kwargs)
-
-  def poll(self, *args, **kwargs):
-    return self
-
-  def wait(self, progress=None):
-    return self
-
-  def kill_threads(self):
-    return self
-
   def __enter__(self):
     return self
 
@@ -572,45 +382,35 @@ class LocalTaskQueue(object):
     self.queue = []
     self.progress = progress
 
-  def insert(self, task, args=[], kwargs={}):
-    # task.payload to ensure conformity with TaskQueue
-    task = {
-      'payload': task.payload(),
-      'id': -1,
-    }
-
-    self.queue.append( (task, args, kwargs) )
-
-  def insert_all(
-      self, tasks, args=[], kwargs={}, 
+  def insert(
+      self, tasks, 
       delay_seconds=0, total=None,
       parallel=None, progress=True
     ):
-
-    if parallel is None:
-      parallel = self.parallel
-
     for task in tasks:
+      args, kwargs = [], {}
+      if isinstance(task, tuple):
+        task, args, kwargs = task
       task = {
         'payload': task.payload(),
         'id': -1,
       }
       self.queue.append( (task, args, kwargs) )
 
-    self._process(progress)
-
-  def wait(self, progress=None):
-    self._process(progress)
-    return self
+  def insert_all(self, *args, **kwargs):
+    self.insert(*args, **kwargs)
+    self.execute(self.progress)
 
   def poll(self, *args, **kwargs):
     pass
 
-  def kill_threads(self):
-    return self
+  def execute(self, progress=True, parallel=None, total=None):
+    if parallel is None:
+      parallel = self.parallel
 
-  def _process(self, progress=True):
-    with tqdm(total=len(self.queue), desc="Tasks", disable=(not progress)) as pbar:
+    total = totalfn(self.queue, total)
+
+    with tqdm(total=total, desc="Tasks", disable=(not progress)) as pbar:
       if self.parallel == 1:
         while self.queue:
           _task_execute(self.queue.pop(0))
@@ -626,7 +426,15 @@ class LocalTaskQueue(object):
     return self
 
   def __exit__(self, exception_type, exception_value, traceback):
-    self._process()
+    self.execute()
+
+class MockTaskQueue(LocalTaskQueue):
+  pass
+
+class GreenTaskQueue(TaskQueue):
+  def __init__(self, *args, **kwargs):
+    kwargs['green'] = True
+    super(GreenTaskQueue, self).__init__(*args, **kwargs)
 
 # Necessary to define here to make the 
 # function picklable
@@ -646,8 +454,8 @@ def _scatter(sequence, n):
   ]
 
 def soloprocess_upload(QueueClass, queue_name, tasks):
-  with QueueClass(queue_name) as tq:
-    tq.insert_all(tasks)
+  tq = QueueClass(queue_name)
+  tq.insert(tasks)
 
 error_queue = mp.Queue()
 
